""" Handy utilities"""
__all__ = ["compute_ramp_factor"]

from elastica.utils import Tolerance


def compute_ramp_factor(time, ramp_interval, start_time, end_time):
    """
    This function returns a linear ramping up factor based on time, ramp_interval,
    start_time and end_time.

    Parameters
    ----------
    time : float
        The time of simulation.
    ramp_interval : float
        ramping time for magnetic field.
    start_time : float
        Turning on time of magnetic field.
    end_time : float
        Turning off time of magnetic field.

    Returns
    -------
    factor : float
        Ramp up factor.

    """
    factor = (time > start_time) * (time <= end_time) * min(
        1.0, (time - start_time) / (ramp_interval + Tolerance.atol())
    ) + (time > end_time) * max(
        0.0, -1 / (ramp_interval + Tolerance.atol()) * (time - end_time) + 1.0
    )
    return factor



__doc__ = """ Module implementation for external magnetic forces for magnetic Cosserat rods."""
__all__ = ["MagneticForces"]

from elastica.external_forces import NoForces
from elastica.rod.cosserat_rod import CosseratRod
from elastica._linalg import _batch_cross, _batch_matvec, _batch_norm
from magneto_pyelastica.magnetic_field import BaseMagneticField
import numpy as np
from typing import Union


class MagneticForces(NoForces):
    """
    This class applies magnetic forces on a magnetic Cosserat rod, based on an
    external magnetic field.

        Attributes
        ----------
        external_magnetic_field: object
            External magnetic field object, that returns the value of the magnetic field vector
            via a .value() method.
        magnetization_collection: np.ndarray
            2D (dim, n_elems) array containing data with 'float' type.
            Magnetization of the rod, defined on the elements, in the material frame.

    """

    def __init__(
        self,
        external_magnetic_field: BaseMagneticField,
        magnetization_density: Union[float, np.ndarray],
        magnetization_direction: np.ndarray,
        rod_volume: np.ndarray,
        rod_director_collection: np.ndarray,
    ):
        """
        Parameters
        ----------
        external_magnetic_field: object
            External magnetic field object, that returns the value of the
            magnetic field vector via a .value() method.
        magnetization_density: float or a np.ndarray
            Float number or 1D (n_elems) array containing data with 'float' type.
            Density of magnetization of the rod.
        magnetization_direction: np.ndarray
            1D (dim) or 2D (dim, n_elems) array containing data with 'float' type.
            Direction of magnetization of the rod in the lab frame.
        rod_volume: numpy.ndarray
            1D (n_elems) array containing data with 'float' type.
            Rod element volumes.
        rod_director_collection: numpy.ndarray
            3D (dim, dim, n_elems) array containing data with 'float' type.
            Array containing rod elemental director matrices.

        """
        super(NoForces, self).__init__()
        self.external_magnetic_field = external_magnetic_field
        rod_n_elem = rod_volume.shape[0]

        # if fixed value, then expand to rod element size
        if magnetization_direction.shape == (3,) or magnetization_direction.shape == (
            3,
            rod_n_elem,
        ):
            magnetization_direction = magnetization_direction.reshape(3, -1) * np.ones(
                (rod_n_elem,)
            )
        else:
            raise ValueError(
                "Invalid magnetization direction! Should be either a (3,) array or "
                "an array of shape (3, num_rod_elements)"
            )
        # normalise for unit vectors
        magnetization_direction /= _batch_norm(magnetization_direction)
        # convert to local frame
        magnetization_direction_in_material_frame = _batch_matvec(
            rod_director_collection, magnetization_direction
        )

        if not (
            isinstance(magnetization_density, float)
            or magnetization_density.shape == (rod_n_elem,)
        ):
            raise ValueError(
                "Invalid magnetization intensity! Should be either a float or "
                "an array of shape (num_rod_elements,)"
            )

        self.magnetization_collection = (
            magnetization_density
            * rod_volume
            * magnetization_direction_in_material_frame
        )

    def apply_torques(self, rod: CosseratRod, time: np.float64 = 0.0):
        rod.external_torques += _batch_cross(
            self.magnetization_collection,
            # convert external_magnetic_field to local frame
            _batch_matvec(
                rod.director_collection,
                self.external_magnetic_field.value(time=time).reshape(
                    3, 1
                )  # broadcasting 3D vector
                * np.ones((rod.n_elems,)),
            ),
        )



__doc__ = """ Module implementation for external magnetic fields for magnetic Cosserat rods."""
__all__ = [
    "BaseMagneticField",
    "ConstantMagneticField",
    "SingleModeOscillatingMagneticField",
]

from magneto_pyelastica.utils import compute_ramp_factor
import numpy as np


class BaseMagneticField:
    """
    This is the base class for external magnetic field objects.

    Notes
    -----
    Every new magnetic field class must be derived
    from BaseMagneticField class.

    """

    def __init__(self):
        """
        BaseMagneticField class does not need any input parameters.
        """

    def value(self, time: np.float64 = 0.0):
        """Returns the value of the magnetic field vector.

        In BaseMagneticField class, this routine simply passes.

        Parameters
        ----------
        time : float
            The time of simulation.

        Returns
        -------

        """


class ConstantMagneticField(BaseMagneticField):
    """
    This class represents a magnetic field constant in time.

        Attributes
        ----------
        magnetic_field_amplitude: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Amplitude of the constant magnetic field.
        ramp_interval : float
            ramping time for magnetic field.
        start_time : float
            Turning on time of magnetic field.
        end_time : float
            Turning off time of magnetic field.

    """

    def __init__(self, magnetic_field_amplitude, ramp_interval, start_time, end_time):
        """

        Parameters
        ----------
        magnetic_field_amplitude: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Amplitude of the constant magnetic field.
        ramp_interval : float
            ramping time for magnetic field.
        start_time : float
            Turning on time of magnetic field.
        end_time : float
            Turning off time of magnetic field.

        """
        super(ConstantMagneticField, self).__init__()
        self.magnetic_field_amplitude = magnetic_field_amplitude
        self.ramp_interval = ramp_interval
        self.start_time = start_time
        self.end_time = end_time

    def value(self, time: np.float64 = 0.0):
        """
        This function returns the value of the magnetic field vector based on the
        magnetic_field_amplitude.

        Parameters
        ----------
        time : float
            The time of simulation.

        Returns
        -------
        magnetic_field: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Value of the constant magnetic field.
        Notes
        -------
        Assumes only time dependence.

        """
        factor = compute_ramp_factor(
            time=time,
            ramp_interval=self.ramp_interval,
            start_time=self.start_time,
            end_time=self.end_time,
        )
        return self.magnetic_field_amplitude * factor


class SingleModeOscillatingMagneticField(BaseMagneticField):
    """
    This class represents a magnetic field oscillating sinusoidally in time
    with one mode.

        Attributes
        ----------
        magnetic_field_amplitude: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Amplitude of the oscillating magnetic field.
        magnetic_field_angular_frequency: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Angular frequency of the oscillating magnetic field.
        magnetic_field_phase_difference: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Phase difference of the oscillating magnetic field.
        ramp_interval : float
            ramping time for magnetic field.
        start_time : float
            Turning on time of magnetic field.
        end_time : float
            Turning off time of magnetic field.

    """

    def __init__(
        self,
        magnetic_field_amplitude,
        magnetic_field_angular_frequency,
        magnetic_field_phase_difference,
        ramp_interval,
        start_time,
        end_time,
    ):
        """

        Parameters
        ----------
        magnetic_field_amplitude: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Amplitude of the oscillating magnetic field.
        magnetic_field_angular_frequency: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Angular frequency of the oscillating magnetic field.
        magnetic_field_phase_difference: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Phase difference of the oscillating magnetic field.
        ramp_interval : float
            ramping time for magnetic field.
        start_time : float
            Turning on time of magnetic field.
        end_time : float
            Turning off time of magnetic field.

        """
        super(SingleModeOscillatingMagneticField, self).__init__()
        self.magnetic_field_amplitude = magnetic_field_amplitude
        self.magnetic_field_angular_frequency = magnetic_field_angular_frequency
        self.magnetic_field_phase_difference = magnetic_field_phase_difference
        self.ramp_interval = ramp_interval
        self.start_time = start_time
        self.end_time = end_time

    def value(self, time: np.float64 = 0.0):
        """
        This function returns the value of the sinusoidally oscillating magnetic field
        vector, based on amplitude, frequency and phase difference.

        Parameters
        ----------
        time : float
            The time of simulation.

        Returns
        -------
        magnetic_field: numpy.ndarray
            1D (dim,) array containing data with 'float' type.
            Value of the oscillatory magnetic field.
        Notes
        -------
        Assumes only time dependence.

        """
        factor = compute_ramp_factor(
            time=time,
            ramp_interval=self.ramp_interval,
            start_time=self.start_time,
            end_time=self.end_time,
        )
        return (
            factor
            * self.magnetic_field_amplitude
            * np.sin(
                self.magnetic_field_angular_frequency * time
                + self.magnetic_field_phase_difference
            )
        )
